package com.woolha.jpamulticolumnselect.repository;

import com.woolha.jpamulticolumnselect.model.Order;
import com.woolha.jpamulticolumnselect.model.OrderLine;
import com.woolha.jpamulticolumnselect.model.OrderLine_;
import com.woolha.jpamulticolumnselect.model.Order_;
import com.woolha.jpamulticolumnselect.model.projection.OrderDto;
import com.woolha.jpamulticolumnselect.model.projection.OrderWithLinesDto;
import jakarta.persistence.EntityManager;
import jakarta.persistence.NoResultException;
import jakarta.persistence.PersistenceContext;
import jakarta.persistence.Tuple;
import jakarta.persistence.criteria.CriteriaBuilder;
import jakarta.persistence.criteria.CriteriaQuery;
import jakarta.persistence.criteria.Join;
import jakarta.persistence.criteria.Root;
import java.math.BigDecimal;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;

public class CustomOrderRepositoryImpl implements CustomOrderRepository {

  @PersistenceContext
  private EntityManager entityManager;

  @Override
  public Optional<OrderDto> findOneProjectedUsingCriteriaQueryById(UUID id) {
    CriteriaBuilder cb = this.entityManager.getCriteriaBuilder();
    CriteriaQuery<Tuple> cq = cb.createTupleQuery();
    Root<Order> root = cq.from(Order.class);

    cq
        .multiselect(
            root.get(Order_.id).alias("id"),
            root.get(Order_.orderNumber).alias("orderNumber"),
            root.get(Order_.totalPrice).alias("totalPrice"),
            root.get(Order_.createdAt).alias("createdAt")
        )
        .where(cb.equal(root.get(Order_.id), id));

    try {
      Tuple tuple = entityManager
          .createQuery(cq)
          .setMaxResults(1)
          .getSingleResult();

      OrderDto orderDto = new OrderDto(
          (UUID) tuple.get("id"),
          (String) tuple.get("orderNumber"),
          (BigDecimal) tuple.get("totalPrice"),
          (ZonedDateTime) tuple.get("createdAt")
      );

      return Optional.of(orderDto);
    } catch (NoResultException ex) {
      return Optional.empty();
    }
  }

  @Override
  public Optional<OrderWithLinesDto> findOneWithLinesProjectedUsingCriteriaQueryById(UUID id) {
    CriteriaBuilder cb = this.entityManager.getCriteriaBuilder();
    CriteriaQuery<Tuple> cq = cb.createTupleQuery();
    Root<Order> root = cq.from(Order.class);
    Join<Order, OrderLine> orderLineJoin = root.join(Order_.orderLines);

    cq
        .multiselect(
            root.get(Order_.id).alias("id"),
            root.get(Order_.orderNumber).alias("orderNumber"),
            root.get(Order_.totalPrice).alias("totalPrice"),
            root.get(Order_.createdAt).alias("createdAt"),
            orderLineJoin.get(OrderLine_.id).alias("orderLineId"),
            orderLineJoin.get(OrderLine_.quantity).alias("orderLineQuantity"),
            orderLineJoin.get(OrderLine_.productName).alias("orderLineProductName"),
            orderLineJoin.get(OrderLine_.price).alias("orderLinePrice")
        )
        .where(cb.equal(root.get(Order_.id), id));

    try {
      List<Tuple> tuples = entityManager
          .createQuery(cq)
          .getResultList();

      List<OrderWithLinesDto.OrderLineDto> orderLines = tuples.stream()
          .map(tuple -> new OrderWithLinesDto.OrderLineDto(
              (UUID) tuple.get("orderLineId"),
              (Integer) tuple.get("orderLineQuantity"),
              (String) tuple.get("orderLineProductName"),
              (BigDecimal) tuple.get("orderLinePrice")
          ))
          .toList();

      OrderWithLinesDto orderWithLinesDto = new OrderWithLinesDto(
          (UUID) tuples.get(0).get("id"),
          (String) tuples.get(0).get("orderNumber"),
          (BigDecimal) tuples.get(0).get("totalPrice"),
          (ZonedDateTime) tuples.get(0).get("createdAt"),
          orderLines
      );

      return Optional.of(orderWithLinesDto);
    } catch (NoResultException ex) {
      return Optional.empty();
    }
  }
}
